/************************************************
 * Copyright (c) IBM Corp. 2014
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *************************************************/

/*
 * Contributors:
 *     arayshu, lschneid - initial implementation
 */

#ifndef __SKV_SERVER_EVENT_TYPE_HPP__
#define __SKV_SERVER_EVENT_TYPE_HPP__

typedef enum
{
  // IT_API related events
  SKV_SERVER_EVENT_TYPE_NONE = 0,
  SKV_SERVER_EVENT_TYPE_IT_UNAFF_ERROR = 1,
  SKV_SERVER_EVENT_IT_ASYNC_UNAFF_SPIGOT_ONLINE,
  SKV_SERVER_EVENT_IT_ASYNC_UNAFF_SPIGOT_OFFLINE,

  SKV_SERVER_EVENT_TYPE_IT_AFF_ERROR,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SEVD_FULL_ERROR,             // 5
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_FAILURE,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_BAD_TRANSPORT_OPCODE,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_REQ_DROPPED,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAW_ACCESS_VIOLATION,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAW_CORRUPT_DATA,          // 10
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAR_ACCESS_VIOLATION,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_LOCAL_ACCESS_VIOLATION,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_RECV_ACCESS_VIOLATION,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_IRRQ_ACCESS_VIOLATION,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_TRANSPORT_ERROR,           // 15
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_LLP_ERROR,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_ERROR,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_ACCESS_VIOLATION,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_RECV_ACCESS_VIOLATION,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_RECV_LENGTH_ERROR,         // 20
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SOFT_HI_WATERMARK,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SRQ_ERROR,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_SRQ_LOW_WATERMARK,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_SRQ_CATASTROPHIC,
  SKV_SERVER_EVENT_IT_ASYNC_AFF_SEVD_FULL_ERROR,                // 25
  SKV_SERVER_EVENT_IT_ASYNC_AFF_SEVD_OP_ERROR,

  SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_BROKEN,
  SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_ACCEPT_ARRIVAL,
  SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_ESTABLISHED,
  SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_DISCONNECT,                 // 30
  SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_PEER_REJECT,

  SKV_SERVER_EVENT_TYPE_IT_CMR_CONN_REQUEST,

  SKV_SERVER_EVENT_TYPE_IT_DTO_SQ_CMPL,
  SKV_SERVER_EVENT_TYPE_IT_DTO_RDMA_WRITE_CMPL,
  SKV_SERVER_EVENT_TYPE_IT_DTO_RDMA_READ_CMPL,                  // 35
  SKV_SERVER_EVENT_TYPE_IT_DTO_INSERT_CMD,
  SKV_SERVER_EVENT_TYPE_IT_DTO_BULK_INSERT_CMD,
  SKV_SERVER_EVENT_TYPE_IT_DTO_UPDATE_CMD,
  SKV_SERVER_EVENT_TYPE_IT_DTO_REMOVE_CMD,
  SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_CMD,                    // 40
  SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_N_KEYS_CMD,
  SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_DIST_CMD,
  SKV_SERVER_EVENT_TYPE_IT_DTO_OPEN_CMD,
  SKV_SERVER_EVENT_TYPE_IT_DTO_PDSCNTL_CMD,

  SKV_SERVER_EVENT_TYPE_ACTIVE_BCAST_CMD,                       // 45
  SKV_SERVER_EVENT_TYPE_CURSOR_PREFETCH_CMD,

  SKV_SERVER_EVENT_TYPE_LOCAL_KV_CMPL,
  SKV_SERVER_EVENT_TYPE_LOCAL_KV_ERROR,
  SKV_SERVER_EVENT_TYPE_LOCAL_KV_ASYNC_ERROR,

// MPI related server_events
} skv_server_event_type_t;

static
const char*
skv_server_event_type_to_string( skv_server_event_type_t aType )
{
  switch( aType )
  {
    case SKV_SERVER_EVENT_TYPE_NONE: { return "SKV_SERVER_EVENT_TYPE_NONE";}
    case SKV_SERVER_EVENT_TYPE_IT_UNAFF_ERROR: { return "SKV_SERVER_EVENT_TYPE_IT_UNAFF_ERROR";}
    case SKV_SERVER_EVENT_IT_ASYNC_UNAFF_SPIGOT_ONLINE: { return "SKV_SERVER_EVENT_IT_ASYNC_UNAFF_SPIGOT_ONLINE";}
    case SKV_SERVER_EVENT_IT_ASYNC_UNAFF_SPIGOT_OFFLINE: { return "SKV_SERVER_EVENT_IT_ASYNC_UNAFF_SPIGOT_OFFLINE";}
    case SKV_SERVER_EVENT_TYPE_IT_AFF_ERROR: { return "SKV_SERVER_EVENT_TYPE_IT_AFF_ERROR";}
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SEVD_FULL_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SEVD_FULL_ERROR"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_FAILURE: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_FAILURE"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_BAD_TRANSPORT_OPCODE: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_BAD_TRANSPORT_OPCODE"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_REQ_DROPPED: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_REQ_DROPPED"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAW_ACCESS_VIOLATION: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAW_ACCESS_VIOLATION"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAW_CORRUPT_DATA:  { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAW_CORRUPT_DATA"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAR_ACCESS_VIOLATION: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_RDMAR_ACCESS_VIOLATION"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_LOCAL_ACCESS_VIOLATION: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_LOCAL_ACCESS_VIOLATION"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_RECV_ACCESS_VIOLATION: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_RECV_ACCESS_VIOLATION"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_IRRQ_ACCESS_VIOLATION: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_IRRQ_ACCESS_VIOLATION"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_TRANSPORT_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_TRANSPORT_ERROR"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_LLP_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_L_LLP_ERROR"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_ERROR"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_ACCESS_VIOLATION: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_ACCESS_VIOLATION"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_RECV_ACCESS_VIOLATION: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_RECV_ACCESS_VIOLATION"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_RECV_LENGTH_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_R_RECV_LENGTH_ERROR"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SOFT_HI_WATERMARK: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SOFT_HI_WATERMARK"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SRQ_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_EP_SRQ_ERROR"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_SRQ_LOW_WATERMARK: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_SRQ_LOW_WATERMARK"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_SRQ_CATASTROPHIC: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_SRQ_CATASTROPHIC"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_SEVD_FULL_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_SEVD_FULL_ERROR"; }
    case SKV_SERVER_EVENT_IT_ASYNC_AFF_SEVD_OP_ERROR: { return "SKV_SERVER_EVENT_IT_ASYNC_AFF_SEVD_OP_ERROR"; }
    case SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_BROKEN: { return "SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_BROKEN";}
    case SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_ACCEPT_ARRIVAL: { return "SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_ACCEPT_ARRIVAL";}
    case SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_ESTABLISHED: { return "SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_ESTABLISHED";}
    case SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_DISCONNECT: { return "SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_DISCONNECT";}
    case SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_PEER_REJECT: { return "SKV_SERVER_EVENT_TYPE_IT_CMM_CONN_PEER_REJECT";}
    case SKV_SERVER_EVENT_TYPE_IT_CMR_CONN_REQUEST: { return "SKV_SERVER_EVENT_TYPE_IT_CMR_CONN_REQUEST";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_RDMA_WRITE_CMPL: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_RDMA_WRITE_CMPL"; }
    case SKV_SERVER_EVENT_TYPE_IT_DTO_RDMA_READ_CMPL: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_RDMA_READ_CMPL"; }
    case SKV_SERVER_EVENT_TYPE_IT_DTO_SQ_CMPL: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_SQ_CMPL";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_INSERT_CMD: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_INSERT_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_BULK_INSERT_CMD: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_BULK_INSERT_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_UPDATE_CMD: { return " SKV_SERVER_EVENT_TYPE_IT_DTO_UPDATE_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_REMOVE_CMD: { return " SKV_SERVER_EVENT_TYPE_IT_DTO_REMOVE_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_CMD: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_N_KEYS_CMD: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_N_KEYS_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_DIST_CMD: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_RETRIEVE_DIST_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_OPEN_CMD: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_OPEN_CMD";}
    case SKV_SERVER_EVENT_TYPE_IT_DTO_PDSCNTL_CMD: { return "SKV_SERVER_EVENT_TYPE_IT_DTO_PDSCNTL_CMD";}
    case SKV_SERVER_EVENT_TYPE_ACTIVE_BCAST_CMD: { return "SKV_SERVER_EVENT_TYPE_ACTIVE_BCAST_CMD";}
    case SKV_SERVER_EVENT_TYPE_CURSOR_PREFETCH_CMD: { return "SKV_SERVER_EVENT_TYPE_CURSOR_PREFETCH_CMD";}
    case SKV_SERVER_EVENT_TYPE_LOCAL_KV_CMPL: { return "SKV_SERVER_EVENT_TYPE_LOCAL_KV_CMPL"; }
    case SKV_SERVER_EVENT_TYPE_LOCAL_KV_ERROR: { return "SKV_SERVER_EVENT_TYPE_LOCAL_KV_ERROR"; }
    case SKV_SERVER_EVENT_TYPE_LOCAL_KV_ASYNC_ERROR: { return "SKV_SERVER_EVENT_TYPE_LOCAL_KV_ASYNC_ERROR"; }
    default:
    {
      StrongAssertLogLine( 0 )
        << "skv_server_event_type_to_string():: ERROR:: "
        << " aType: " << aType
        << " Not supported"
        << EndLogLine;
        return "SKV_SERVER_EVENT_TYPE_UNKNOWN";
    }
  }
}

#endif
